# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/Integration_Test_Generator.ipynb.

# %% auto 0
__all__ = ['create_venv_and_run_tests_bash_script', 'fix_requirements_and_run_tests']

# %% ../../nbs/Integration_Test_Generator.ipynb 1
from typing import *
import shutil
from pathlib import Path
import os
import toml
import re
from tempfile import TemporaryDirectory
from contextlib import contextmanager
from collections import defaultdict
import subprocess  # nosec: B404: Consider possible security implications associated with the subprocess module.

from yaspin import yaspin

from .logger import get_logger

from .._code_generator.prompts import REQUIREMENTS_GENERATION_PROMPT
from faststream_gen._code_generator.constants import (
    APPLICATION_FILE_PATH,
    TEST_FILE_PATH,
    STEP_LOG_DIR_NAMES,
    TOML_FILE_NAME,
    OpenAIModel,
)

from .._code_generator.chat import CustomAIChat, ValidateAndFixResponse

from faststream_gen._code_generator.helper import (
    download_and_extract_faststream_archive,
    write_file_contents,
    read_file_contents,
    set_cwd,
    mock_openai_create,
    retry_on_error,
)

# %% ../../nbs/Integration_Test_Generator.ipynb 3
create_venv_and_run_tests_bash_script = """
#!/bin/bash

# get the project directory from command line argument
project_dir=$1

# get the venv directory from command line argument
venv_dir=$2

# create a new venv in the specified directory
python3 -m venv $venv_dir/my_venv > /dev/null 2>&1

# activate the venv
source $venv_dir/my_venv/bin/activate

# navigate to the project directory
cd $project_dir

# install the python project inside the venv
pip install .['dev'] > /dev/null 2>&1

# run pytest and capture output
pytest_output=$(pytest --tb=short)

# capture pytest exit code
pytest_exit_code=$?

# print the pytest output
echo "pytest_output_start:$pytest_output:pytest_output_end"

# print the pytest exit code
echo "pytest_exit_code:$pytest_exit_code"

# deactivate the venv
deactivate
"""

# %% ../../nbs/Integration_Test_Generator.ipynb 4
def _setup_venv_and_run_tests(output_path: str) -> List[str]:
    output_path_resolved = Path(output_path).resolve()
    with TemporaryDirectory() as d:
        bash_file = Path(d) / "run_tests.sh"
        write_file_contents(str(bash_file), create_venv_and_run_tests_bash_script)
        with set_cwd(d):
            # nosemgrep: python.lang.security.audit.subprocess-shell-true.subprocess-shell-true
            p = subprocess.run( # nosec: B602, B603, B607 subprocess call - check for execution of untrusted input.
                ["bash", "run_tests.sh", output_path_resolved, Path(d).resolve()],
                capture_output=True,
                text=True,
            )
            
        # Extract exit code
        exit_code = int(re.search('pytest_exit_code:(\d+)', p.stdout).group(1)) # type: ignore
        if exit_code !=0:
            # Extract pytest output
            pytest_output: str = re.search('pytest_output_start:(.*):pytest_output_end', p.stdout, re.DOTALL).group(1).strip() # type: ignore
            return [pytest_output]
        
        return []    

# %% ../../nbs/Integration_Test_Generator.ipynb 7
def _stript(s: str) -> str:
    return s.strip().strip('"')


def _split_app_and_test_req(response: str) -> Tuple[str, str]:

    app_req, test_req = response.split("==== APP REQUIREMENT ====")[1].split(
        "==== TEST REQUIREMENT ===="
    )
    return _stript(app_req), _stript(test_req)

# %% ../../nbs/Integration_Test_Generator.ipynb 10
def _update_toml_file(output_dir: str, app_req: str, test_req: str) -> None:
    toml_file_path = f"{output_dir}/{TOML_FILE_NAME}"
    toml_contents = read_file_contents(toml_file_path)
    data = toml.loads(toml_contents)
    
    app_reqs = [r.strip() for r in app_req.split(",")]
    test_reqs = [r.strip() for r in test_req.split(",")]
    test_reqs = [r for r in test_reqs if r != "pytest"]
    
    data["project"]["dependencies"] = data["project"]["dependencies"] + app_reqs
    data["project"]["optional-dependencies"]["testing"] = data["project"]["optional-dependencies"]["testing"] + test_reqs
    
    toml_string = toml.dumps(data)
    write_file_contents(toml_file_path, toml_string)

# %% ../../nbs/Integration_Test_Generator.ipynb 12
def _validate_response(
    response: str, output_directory: str, **kwargs: Dict[str, Any]
) -> List[str]:
    try:
        app_req, test_req = _split_app_and_test_req(response)
    except (IndexError, ValueError) as e:
        return [
            "Please add ==== APP REQUIREMENT ==== and ==== TEST REQUIREMENT ==== in your response"
        ]
    
    _update_toml_file(output_directory, app_req, test_req)
    
    return _setup_venv_and_run_tests(output_directory)

# %% ../../nbs/Integration_Test_Generator.ipynb 16
@retry_on_error()  # type: ignore
def _generate(
    model: str,
    prompt: str,
    app_and_test_code: str,
    total_usage: List[Dict[str, int]],
    output_directory: str,
    **kwargs,
) -> Tuple[str, List[Dict[str, int]], bool]:
    requirements_generator = CustomAIChat(
        params={
            "temperature": 0.2,
        },
        model=model,
        user_prompt=prompt,
    )
    requirements_validator = ValidateAndFixResponse(
        requirements_generator, _validate_response
    )
    validator_result = requirements_validator.fix(
        app_and_test_code,
        total_usage,
        STEP_LOG_DIR_NAMES["requirements"],
        str(output_directory),
        **kwargs,
    )
    return (
        (validator_result, True) # type: ignore
        if isinstance(validator_result[-1], defaultdict)
        else validator_result
    )

# %% ../../nbs/Integration_Test_Generator.ipynb 19
def fix_requirements_and_run_tests(
    output_directory: str,
    model: str,
    total_usage: List[Dict[str, int]],
) -> Tuple[List[Dict[str, int]], bool]:
    with yaspin(
        text="Running integration tests...", color="cyan", spinner="clock"
    ) as sp:
        app_file = Path(output_directory) / APPLICATION_FILE_PATH
        app_code = read_file_contents(str(app_file))

        test_file = Path(output_directory) / TEST_FILE_PATH
        test_code = read_file_contents(str(test_file))

        app_and_test_code = f"==== APP CODE ====\n\n{app_code}\n\n==== TEST CODE ====\n\n{test_code}\n\n"

        total_usage, is_requirements_file_valid = _generate(
            model,
            REQUIREMENTS_GENERATION_PROMPT,
            app_and_test_code,
            total_usage,
            output_directory,
        )

        sp.text = ""
        if is_requirements_file_valid:
            message = " ✔ Integration tests were successfully completed."
        else:
            message = " ✘ Error: Integration tests failed."
            sp.color = "red"

        sp.ok(message)

        return total_usage, is_requirements_file_valid



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/PackageData.ipynb.

# %% auto 0
__all__ = ['get_root_data_path']

# %% ../../nbs/PackageData.ipynb 1
from pathlib import Path
from typing import *

# %% ../../nbs/PackageData.ipynb 2
def get_root_data_path() -> Path:
    """Returns path to package_data in lib

    Example:
        ``` python
        from nbdev_mkdocs._package_data import get_root_data_path

        mkdocs_template_path = get_root_data_path() / "mkdocs_template.yml"
        print(f"Path is: {mkdocs_template_path.resolve()}")
        assert mkdocs_template_path.exists()

        ```

    !!! note

        The above docstring is autogenerated by docstring-gen library (https://github.com/airtai/docstring-gen)
    """
    return Path(__file__).parent.parent / "package_data"



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/Prompts.ipynb.

# %% auto 0
__all__ = ['SYSTEM_PROMPT', 'APP_VALIDATION_PROMPT', 'APP_SKELETON_GENERATION_PROMPT', 'APP_AND_TEST_GENERATION_PROMPT',
           'REQUIREMENTS_GENERATION_PROMPT']

# %% ../../nbs/Prompts.ipynb 1
SYSTEM_PROMPT = """
You are an expert Python developer, tasked to generate executable Python code as a part of your work with the FastStream framework. 

You are to abide by the following guidelines:

1. You must never enclose the generated Python code with ``` python. It is mandatory that the output is a valid and executable Python code. Please ensure this rule is never broken.

2. Some prompts might require you to generate code that contains async functions. For example:

async def app_setup(context: ContextRepo):
    raise NotImplementedError()

In such cases, it is necessary to add the "import asyncio" statement at the top of the code. 

3. Whenever you are creating a message class while generating Faststream skeleton and the application code, make sure the message class is a derived class of BaseModel from pydantic.

        Example of a Valid message class:
            class Pet(BaseModel):
                pet_id: NonNegativeInt = Field(..., examples=[1], description="Int data example")
                species: str = Field(..., examples=["dog"], description="Pet example")

        Example of a invalid message class:
            class Pet:
                def __init__(self, pet_id: int, species: str):
                    self.pet_id = pet_id
                    self.species = species
                    
4. When generating a lists of external dependencies from both the ==== APP CODE ==== and ==== TEST CODE ==== sections, include only external libraries and not internal Python libraries like json, time, asyncio, etc., in the ==== APP REQUIREMENT ==== or ==== TEST REQUIREMENT ==== sections. Additionally do not include pytest in the ==== TEST REQUIREMENT ====

You will encounter sections marked as:

==== APP DESCRIPTION: ====

These sections contain the description of the FastStream app you need to implement. Treat everything below this line, until the end of the prompt, as the description to follow for the app implementation.
"""

# %% ../../nbs/Prompts.ipynb 2
APP_VALIDATION_PROMPT = """
You should provide a response of 0, 1, 2, or 3, and nothing else, based on the following rules:

==== RULES: ====

If the ==== APP DESCRIPTION: ==== section is not related to FastStream or contains violence, self-harm, harassment/threatening, or hate/threatening information, respond with 0.

If the ==== APP DESCRIPTION: ==== section is related to FastStream but primarily provides general information about FastStream and what it is, respond with 1.

If it is NOT possible to infer the topic name or there is no explanation about the business logic in the ==== APP DESCRIPTION: ==== section, respond with 2. This is crucial.

If the ==== APP DESCRIPTION: ==== section is related to FastStream, provides instructions on which topic the messages should be consumed/produced, and includes at least one defined topic, respond with 3.

Here are few examples for your understanding:


==== EXAMPLE APP DESCRIPTION ====
Generate a new FastStream app, which has a producer function and a consumer function 
==== YOUR RESPONSE ====
2

==== EXAMPLE APP DESCRIPTION ====
In App description 1, the user has not defined the message structure or the topic name to publish/subscribe. As a result, you should respond with 2. 
==== YOUR RESPONSE ====
2

==== EXAMPLE APP DESCRIPTION ====
Create a FastStream application.
==== YOUR RESPONSE ====
2

==== EXAMPLE APP DESCRIPTION ====
create FastStream app where message has user_data attribute.
==== YOUR RESPONSE ====
2

==== EXAMPLE APP DESCRIPTION ====
FastStream app with for consuming messages from the hello topic
==== YOUR RESPONSE ====
3

==== EXAMPLE APP DESCRIPTION ====
Write a FastStream application with with one consumer function and two producer functions. The consumer function should receive the a message posted on "new_joinee" topic. The message should contain "employee_name", "age", "location" and "experience" attributes. After consuming the consumer function should send the details to the "project_team" and "admin_team" topics. Use only localhost broker==== Response 5 ====
==== YOUR RESPONSE ====
3

==== EXAMPLE APP DESCRIPTION ====
Develop a new FastStream application that consumes JSON-encoded objects from the "receive_order" topic. These objects include attributes like "name" and "quantity." Upon consumption, enhance the message by adding a "location" attribute set to "Zagreb." Subsequently, forward the modified message to the "place_order" topic. After this, send another message to the "update_inventory" topic. This message should include a "quantity" attribute that corresponds to the received quantity value. No authentication is required.==== Response 6 ====
==== YOUR RESPONSE ====
3

==== EXAMPLE APP DESCRIPTION ====
Who are you
==== YOUR RESPONSE ====
0

==== EXAMPLE APP DESCRIPTION ====
What is the latest vesion of FastStream
==== YOUR RESPONSE ====
1

Please respond only with numbers 0, 1, 2 or 3 (WITH NO ADDITIONAL TEXT!)

==== APP DESCRIPTION: ====

"""

# %% ../../nbs/Prompts.ipynb 3
APP_SKELETON_GENERATION_PROMPT = """


Generate skeleton code for FastStream applications based on provided app descriptions in the "==== USER APP DESCRIPTION ====" section, adhering to these guidelines:

    - Avoid implementing business logic of ANY function. Instead, write "raise NotImplementedError()" and create Google-style docstrings to describe their intended functionality when handling received or produced messages. In each docstring, include a clear instruction to "log the consumed message using logger.info" for subscriber functions.

    - Ensure the generated code aligns with the specific app description requirements.

    - Provide a clear and organized starting point for developers.

    - Your response must contain only valid Python code, saveable as a .py script; no additional text is allowed.
    
    - DO NOT enclose the response within back-ticks. Meaning NEVER ADD ```python to your response.


The goal is to offer developers a structured starting point that matches the example app descriptions, aiding in FastStream application development. Follow the example patterns provided for reference.


==== RELEVANT EXAMPLES GOES HERE ====


==== USER APP DESCRIPTION: ====

"""

# %% ../../nbs/Prompts.ipynb 4
APP_AND_TEST_GENERATION_PROMPT = """
You will be provided with a description about FastStream application in ==== APP DESCRIPTION ==== section and the skeleton code in ==== APP SKELETON ==== section. Your goal is to generate both the application and the test code based on the provided ==== APP SKELETON ==== section and the description ==== APP DESCRIPTION ==== in a single Python file. Your response will be split the application.py and the test.py files and saved to the disc.

Input:

You will be given skeleton code and the description for a FastStream application in ==== APP SKELETON ==== section and ==== APP DESCRIPTION ==== section respectively. The skeleton code and the description will have implementation details.

Output:

You need to understand the business logic mentioned in in ==== APP SKELETON ==== section and ==== APP DESCRIPTION ==== sections and generate the following:

    - The application code by implementing the methods decorated with @broker.subscriber and @broker.publisher in the docstring
    - The test code which tests the functionality of the generated application file

You need to generate a single valid Python file containing both the application and the test code. Remember that the application and the test code will be split into two and saved to the disc as application.py and test.py. So while writing the test code, make sure to import all the necessary symbols from the application.py

Application Code:

Generate the entire application code from the provided skeleton. You must implement all of the business logic specified in the docstrings of the @broker.subscriber and @broker.publisher decorated functions in the skeleton code provided in the "==== APP SKELETON ====" section. When implementing business logic for methods decorated with @broker.subscriber and @broker.publisher, strictly follow the instructions in the docstring. Other parts of the code should not be changed.

Test Code:

Create test code using the TestBroker context managers to validate the functionality of the application code.

The FastStream apps can be tested using the TestBroker context managers which, by default, puts the Broker into "testing mode". The Tester will redirect your subscriber and publisher decorated functions to the InMemory brokers so that you can quickly test your app without the need for a running broker and all its dependencies.

While generating the application and the test code for FastStream application based on provided application code in the ==== APP SKELETON: ==== section, adhering to these guidelines:

    - You need to generate the app code and the test code for the application code mentioned in ==== APP SKELETON: ==== in a single valid python file
    - Follow the PEP 8 Style Guide for Python while writing the code
    - Write optimised and readable Code
    - Output only a valid executable python code. No other extra text should be included in your response.
    - DO NOT enclose the response within back-ticks. Meaning NEVER ADD ```python to your response.
    - Never try to explain and reason your answers. Only return a valid python code.
    - Your response should be divided into two section, ### application.py ### which contains the application code and ### test.py ### which contains the test code.
    - Remember, your response must have only two seperators ### application.py ### and ### test.py ###. DO NOT add any other separators (e.g. ### main.py ###).
    - NEVER initialize publisher variable inside functions. You can use broker.publish function only in the test functions. In ALL OTHER functions, use publisher!
    - ALWAYS initialize publisher variable (If you need to publish message) after defining app variable:

        Example 1:
        app = FastStream(broker)
        publisher = broker.publisher("new_data")

        Example 2:
        app = FastStream(broker)
        new_data_publisher = broker.publisher("new_data")

Guidelines for writing tests: 
    - When a function is decorated @broker.publisher in the application code, remember that this function will always publish the message irrespective of the conditions mentioned in the "==== APP SKELETON ====" section. In such cases, never use mock.assert_not_called() while testing the function. For example:
    - When you test a function in the application which is decorated with @broker.publisher("output_data"), you should never use on_output_data.mock.assert_not_called() in your tests. Never ever break this rule. Because the decorator @broker.publisher will always publish the message irrespective of the conditions mentioned in the decorated function.

Below are few examples for your understanding:

==== RELEVANT EXAMPLES GOES HERE ====

You need to avoid the below common issues while generating the test code. The potential fixes for the common issues are also given for your reference. You should use this information as an additional reference while generating the test:

==== Error ====

AssertionError: Expected 'mock' to not have been called. Called 2 times.

==== ERROR CODE ====
async with TestKafkaBroker(broker):
await broker.publish(DataBasic(data=0.2), "input_data")

        on_input_data.mock.assert_called_with(dict(DataBasic(data=0.2)))
        on_output_data.mock.assert_not_called() # ERROR IN THIS LINE

==== FIXED CODE ====

    async with TestKafkaBroker(broker):
        await broker.publish(DataBasic(data=0.2), "input_data")

        on_input_data.mock.assert_called_with(dict(DataBasic(data=0.2)))
        on_output_data.mock.assert_called_once_with(dict(DataBasic(data=1.2))) # ERROR FIXED IN THIS LINE

==== Error ====

pydantic_core.\_pydantic_core.ValidationError: 1 validation error for ResponseModel

==== ERROR CODE ====
@broker.subscriber("output_data")
async def on_output_data(msg: DataBasic) -> DataBasic: # ERROR IN THIS LINE
pass

==== FIXED CODE ====

@broker.subscriber("output_data")
async def on_output_data(msg: DataBasic): # ERROR FIXED IN THIS LINE
pass

==== Error ====

ImportError: cannot import name 'on_output_data' from 'app.application'

==== ERROR CODE ====

from app.application import InputData, broker, on_input_data, on_output_data

==== FIXED CODE ====

from app.application import InputData, broker, on_input_data

==== Error ====

assert on_input_data.mock.assert_called_with(dict(DataMessage(data=1)))

==== ERROR CODE ====

assert on_input_data.mock.assert_called_with(dict(DataMessage(data=1)))

==== FIXED CODE ====

on_input_data.mock.assert_called_with(dict(DataMessage(data=1)))

==== APP DESCRIPTION ====

==== REPLACE WITH APP DESCRIPTION ====


==== APP SKELETON ====

"""

# %% ../../nbs/Prompts.ipynb 5
REQUIREMENTS_GENERATION_PROMPT = """

You are required to examine the Python files labeled as ==== APP CODE ==== and ==== TEST CODE ==== thoroughly. Your task is to create separate lists of external dependencies for both the ==== APP CODE ==== and ==== TEST CODE ==== sections, following the specified format:

==== APP REQUIREMENT ====

==== TEST REQUIREMENT ====

Please adhere to the following rules when compiling the lists of external dependencies:

- Do not include faststream and faststream.kafka in either the ==== APP REQUIREMENT ==== or ==== TEST REQUIREMENT ==== sections. These are considered internal packages and should not be listed as external dependencies.
- Ensure that you populate the ==== APP REQUIREMENT ==== section based on the dependencies found in the ==== APP CODE ==== and the ==== TEST REQUIREMENT ==== section based on the dependencies found in the ==== TEST CODE ====—do not mix them together.
- Include only external libraries and not internal Python libraries like json, time, asyncio, etc., in the ==== APP REQUIREMENT ==== or ==== TEST REQUIREMENT ==== sections.
- Additionally do not include pytest in the ==== TEST REQUIREMENT ====

Below are few examples for your understanding:

==== EXAMPLE APP CODE ==== 

from datetime import datetime
from typing import Optional
import asyncio

from pydantic import BaseModel, Field

from faststream import FastStream, Logger
from faststream.kafka import KafkaBroker
import json


class CourseUpdates(BaseModel):
    course_name: str = Field(..., examples=["Biology"], description="Course example")
    new_content: Optional[str] = Field(
        default=None, examples=["New content"], description="Content example"
    )
    timestamp: datetime = Field(
        ...,
        examples=["2020-04-23 10:20:30.400000"],
        description="The timestamp of the record",
    )


broker = KafkaBroker("localhost:9092")
app = FastStream(broker)


@broker.publisher("notify_updates")
@broker.subscriber("course_updates")
async def on_course_update(msg: CourseUpdates, logger: Logger) -> CourseUpdates:
    logger.info(msg)

    if msg.new_content:
        logger.info(f"Course has new content {msg.new_content=}")
        msg = CourseUpdates(
            course_name=("Updated: " + msg.course_name),
            new_content=msg.new_content,
            timestamp=datetime.now(),
        )
    return msg


==== EXAMPLE TEST CODE ====

from datetime import datetime

import pytest
from freezegun import freeze_time

from faststream._compat import model_to_jsonable
from faststream.kafka import TestKafkaBroker

from .app import CourseUpdates, broker, on_course_update


@broker.subscriber("notify_updates")
async def on_notify_update(msg: CourseUpdates):
    pass


# Feeze time so the datetime always uses the same time
@freeze_time("2023-01-01")
@pytest.mark.asyncio
async def test_app_without_new_content():
    async with TestKafkaBroker(broker):
        timestamp = datetime.now()
        await broker.publish(
            CourseUpdates(course_name="Biology", timestamp=timestamp), "course_updates"
        )

        course_json = model_to_jsonable(
            CourseUpdates(course_name="Biology", timestamp=timestamp)
        )
        on_course_update.mock.assert_called_with(course_json)
        on_notify_update.mock.assert_called_with(course_json)


# Feeze time so the datetime always uses the same time
@freeze_time("2023-01-01")
@pytest.mark.asyncio
async def test_app_with_new_content():
    async with TestKafkaBroker(broker):
        timestamp = datetime.now()
        await broker.publish(
            CourseUpdates(
                course_name="Biology",
                new_content="We have additional classes...",
                timestamp=timestamp,
            ),
            "course_updates",
        )
        course_json = model_to_jsonable(
            CourseUpdates(
                course_name="Biology",
                new_content="We have additional classes...",
                timestamp=timestamp,
            )
        )
        on_course_update.mock.assert_called_with(course_json)

        on_update_json = model_to_jsonable(
            CourseUpdates(
                course_name="Updated: Biology",
                new_content="We have additional classes...",
                timestamp=timestamp,
            )
        )
        on_notify_update.mock.assert_called_with(on_update_json)

==== YOUR RESPONSE ====

==== APP REQUIREMENT ====
"pydantic"

==== TEST REQUIREMENT ====
"freezegun"


==== EXAMPLE APP CODE ====

import asyncio
import json
from datetime import datetime

import requests
from pydantic import BaseModel, Field, NonNegativeFloat

from faststream import ContextRepo, FastStream, Logger
from faststream.kafka import KafkaBroker

broker = KafkaBroker("localhost:9092")
app = FastStream(broker)


publisher = broker.publisher("weather")


class Weather(BaseModel):
    latitude: NonNegativeFloat = Field(
        ...,
        examples=[22.5],
        description="Latitude measures the distance north or south of the equator.",
    )
    longitude: NonNegativeFloat = Field(
        ...,
        examples=[55],
        description="Longitude measures distance east or west of the prime meridian.",
    )
    temperature: float = Field(
        ..., examples=[20], description="Temperature in Celsius degrees"
    )
    windspeed: NonNegativeFloat = Field(
        ..., examples=[20], description="Wind speed in kilometers per hour"
    )
    time: str = Field(
        ..., examples=["2023-09-13T07:00"], description="The time of the day"
    )


@app.on_startup
async def app_setup(context: ContextRepo):
    context.set_global("app_is_running", True)


@app.on_shutdown
async def shutdown(context: ContextRepo):
    context.set_global("app_is_running", False)

    # Get all the running tasks and wait them to finish
    publish_tasks = context.get("publish_tasks")
    await asyncio.gather(*publish_tasks)


async def fetch_and_publish_weather(
    latitude: float,
    longitude: float,
    logger: Logger,
    context: ContextRepo,
    time_interval: int = 5,
) -> None:
    # Always use context: ContextRepo for storing app_is_running variable
    while context.get("app_is_running"):
        uri = f"https://api.open-meteo.com/v1/forecast?current_weather=true&latitude={latitude}&longitude={longitude}"
        response = requests.get(uri)

        if response.status_code == 200:
            # read json response
            raw_data = json.loads(response.content)
            temperature = raw_data["current_weather"]["temperature"]
            windspeed = raw_data["current_weather"]["windspeed"]
            time = raw_data["current_weather"]["time"]

            new_data = Weather(
                latitude=latitude,
                longitude=longitude,
                temperature=temperature,
                windspeed=windspeed,
                time=time,
            )
            key = str(latitude) + "_" + str(longitude)
            await publisher.publish(new_data, key=key.encode("utf-8"))
        else:
            logger.warning(f"Failed API request {uri} at time {datetime.now()}")
        await asyncio.sleep(time_interval)


@app.after_startup
async def publish_weather(logger: Logger, context: ContextRepo):
    logger.info("Starting publishing:")

    latitudes = [13, 50, 44, 24]
    longitudes = [17, 13, 45, 70]
    # start scraping and producing to kafka topic
    publish_tasks = [
        asyncio.create_task(
            fetch_and_publish_weather(latitude, longitude, logger, context)
        )
        for latitude, longitude in zip(latitudes, longitudes)
    ]
    # you need to save asyncio tasks so you can wait them to finish at app shutdown (the function with @app.on_shutdown function)
    context.set_global("publish_tasks", publish_tasks)

==== EXAMPLE TEST CODE ====

import pytest

from faststream import Context, TestApp
from faststream.kafka import TestKafkaBroker

from .app import Weather, app, broker


@broker.subscriber("weather")
async def on_weather(msg: Weather, key: bytes = Context("message.raw_message.key")):
    pass


@pytest.mark.asyncio
async def test_message_was_published():
    async with TestKafkaBroker(broker):
        async with TestApp(app):
            await on_weather.wait_call(3)
            on_weather.mock.assert_called()

==== YOUR RESPONSE ====

==== APP REQUIREMENT ====
"requests, pydantic"

==== TEST REQUIREMENT ====


"""



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/App_Skeleton_Generator.ipynb.

# %% auto 0
__all__ = ['logger', 'generate_app_skeleton']

# %% ../../nbs/App_Skeleton_Generator.ipynb 1
from typing import *
import time
import json
from pathlib import Path
from collections import defaultdict

from yaspin import yaspin

from .._components.logger import get_logger
from .chat import CustomAIChat, ValidateAndFixResponse
from faststream_gen._code_generator.helper import (
    write_file_contents,
    validate_python_code,
    retry_on_error,
)
from .prompts import APP_SKELETON_GENERATION_PROMPT
from faststream_gen._code_generator.constants import (
    STEP_LOG_DIR_NAMES,
    APPLICATION_FILE_PATH,
    LOGS_DIR_NAME
)

# %% ../../nbs/App_Skeleton_Generator.ipynb 3
logger = get_logger(__name__)

# %% ../../nbs/App_Skeleton_Generator.ipynb 5
def _validate_response(
    response: str, output_directory: str, **kwargs: Dict[str, Any]
) -> List[str]:
    target_file_name = Path(output_directory) / APPLICATION_FILE_PATH
    write_file_contents(str(target_file_name), response)
    ret_val = validate_python_code(str(target_file_name))

    return ret_val

# %% ../../nbs/App_Skeleton_Generator.ipynb 8
@retry_on_error()  # type: ignore
def _generate(
    model: str,
    prompt: str,
    app_description_content: str,
    total_usage: List[Dict[str, int]],
    output_directory: str,
    **kwargs,
) -> Tuple[str, List[Dict[str, int]]]:
    app_generator = CustomAIChat(
        params={
            "temperature": 0.2,
        },
        model=model,
        user_prompt=prompt,
        #         semantic_search_query=app_description_content,
    )
    app_validator = ValidateAndFixResponse(app_generator, _validate_response)
    validator_result = app_validator.fix(
        app_description_content,
        total_usage,
        STEP_LOG_DIR_NAMES["skeleton"],
        str(output_directory),
        **kwargs,
    )

    return (
        (validator_result, True)
        if isinstance(validator_result[-1], defaultdict)
        else validator_result
    )

# %% ../../nbs/App_Skeleton_Generator.ipynb 11
def generate_app_skeleton(
    validated_description: str,
    output_directory: str,
    model: str,
    total_usage: List[Dict[str, int]],
    relevant_prompt_examples: str,
) -> Tuple[List[Dict[str, int]], bool]:
    """Generate skeleton code for the new FastStream app from the application description

    Args:
        code_gen_directory: The directory containing the generated files.
        total_usage: list of token usage.
        relevant_prompt_examples: Relevant examples to add in the prompts.

    Returns:
        The total token used to generate the FastStream code
    """
    logger.info("==== Description to Skeleton Generation ====")
    with yaspin(
        text=f"Generating FastStream app skeleton code (usually takes around 15 to 45 seconds)...",
        color="cyan",
        spinner="clock",
    ) as sp:
        prompt = APP_SKELETON_GENERATION_PROMPT.replace(
            "==== RELEVANT EXAMPLES GOES HERE ====", f"\n{relevant_prompt_examples}"
        )

        total_usage, is_valid_skeleton_code = _generate(
            model, prompt, validated_description, total_usage, output_directory
        )
        sp.text = ""
        if is_valid_skeleton_code:
            message = f" ✔ FastStream app skeleton code generated."
        else:
            message = " ✘ Error: Failed to generate a valid application skeleton code."
            sp.color = "red"

        sp.ok(message)
        return total_usage, is_valid_skeleton_code



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/New_Project_Generator.ipynb.

# %% auto 0
__all__ = ['create_project']

# %% ../../nbs/New_Project_Generator.ipynb 1
from typing import *
from pathlib import Path
from yaspin import yaspin
import shutil
import os
import stat


from faststream_gen._code_generator.helper import (
    download_and_extract_faststream_archive,
)
from faststream_gen._code_generator.constants import (
    FASTSTREAM_TEMPLATE_ZIP_URL,
    FASTSTREAM_TEMPLATE_DIR_SUFFIX,
    APPLICATION_FILE_PATH,
    TEST_FILE_PATH,
)

from .._code_generator.helper import write_file_contents

# %% ../../nbs/New_Project_Generator.ipynb 3
_STAT_0o775 = ( stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR
             | stat.S_IRGRP | stat.S_IWGRP | stat.S_IXGRP
             | stat.S_IROTH |                stat.S_IXOTH )


def create_project(
    output_path: str,
) -> None:
    with yaspin(
        text="Creating a new FastStream project...", color="cyan", spinner="clock"
    ) as sp:
        with download_and_extract_faststream_archive(
            FASTSTREAM_TEMPLATE_ZIP_URL
        ) as extracted_path:
            app_file = str(
                extracted_path / FASTSTREAM_TEMPLATE_DIR_SUFFIX / APPLICATION_FILE_PATH
            )
            write_file_contents(app_file, "")
            test_file = str(
                extracted_path / FASTSTREAM_TEMPLATE_DIR_SUFFIX / TEST_FILE_PATH
            )
            write_file_contents(test_file, "")

            for p in (
                Path(extracted_path) / FASTSTREAM_TEMPLATE_DIR_SUFFIX / "scripts"
            ).glob("*.sh"):
                p.chmod(_STAT_0o775)

            shutil.copytree(
                str(extracted_path / FASTSTREAM_TEMPLATE_DIR_SUFFIX),
                output_path,
                dirs_exist_ok=True,
            )

        sp.text = ""
        sp.ok(f" ✔ New FastStream project created.")



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/Chat.ipynb.

# %% auto 0
__all__ = ['logger', 'CustomAIChat', 'ValidateAndFixResponse']

# %% ../../nbs/Chat.ipynb 1
from typing import *
import random
import logging
import time
from collections import defaultdict
from pathlib import Path

import openai
from fastcore.foundation import patch
from langchain.schema.document import Document
from langchain.embeddings import OpenAIEmbeddings
from langchain.vectorstores import FAISS

from faststream_gen._code_generator.constants import (
    DEFAULT_PARAMS,
    MAX_RETRIES,
    STEP_LOG_DIR_NAMES,
    MAX_NUM_FIXES_MSG,
    INCOMPLETE_DESCRIPTION,
    DESCRIPTION_EXAMPLE,
    LOGS_DIR_NAME,
)
from .._components.logger import get_logger, set_level
from .prompts import SYSTEM_PROMPT
from .helper import add_tokens_usage
from .._components.package_data import get_root_data_path

# %% ../../nbs/Chat.ipynb 3
logger = get_logger(__name__, level=logging.WARNING)

# %% ../../nbs/Chat.ipynb 5
# Reference: https://github.com/openai/openai-cookbook/blob/main/examples/How_to_handle_rate_limits.ipynb


def _retry_with_exponential_backoff(
    initial_delay: float = 1,
    exponential_base: float = 2,
    jitter: bool = True,
    max_retries: int = 10,
    max_wait: float = 60,
    errors: tuple = (
        openai.error.RateLimitError,
        openai.error.ServiceUnavailableError,
        openai.error.APIError,
    ),
) -> Callable:
    """Retry a function with exponential backoff."""

    def decorator(
        func: Callable[[str], Tuple[str, str]]
    ) -> Callable[[str], Tuple[str, str]]:
        def wrapper(*args, **kwargs):  # type: ignore
            num_retries = 0
            delay = initial_delay

            while True:
                try:
                    return func(*args, **kwargs)

                except errors as e:
                    num_retries += 1
                    if num_retries > max_retries:
                        raise Exception(
                            f"Maximum number of retries ({max_retries}) exceeded."
                        )
                    delay = min(
                        delay
                        * exponential_base
                        * (1 + jitter * random.random()),  # nosec
                        max_wait,
                    )
                    logger.info(
                        f"Note: OpenAI's API rate limit reached. Command will automatically retry in {int(delay)} seconds. For more information visit: https://help.openai.com/en/articles/5955598-is-api-usage-subject-to-any-rate-limits",
                    )
                    time.sleep(delay)

                except Exception as e:
                    raise e

        return wrapper

    return decorator

# %% ../../nbs/Chat.ipynb 8
def _get_relevant_document(query: str) -> str:
    """Load the vector database and retrieve the most relevant document based on the given query.

    Args:
        query: The query for relevance-based document retrieval.

    Returns:
        The content of the most relevant document as a string.
    """
    db_path = get_root_data_path() / "docs"
    db = FAISS.load_local(db_path, OpenAIEmbeddings()) # type: ignore
    results = db.max_marginal_relevance_search(query, k=1, fetch_k=3)
    results_str = "\n".join([result.page_content for result in results])
    return results_str

# %% ../../nbs/Chat.ipynb 10
class CustomAIChat:
    """Custom class for interacting with OpenAI

    Attributes:
        model: The OpenAI model to use. If not passed, defaults to gpt-3.5-turbo-16k.
        system_prompt: Initial system prompt to the AI model. If not passed, defaults to SYSTEM_PROMPT.
        initial_user_prompt: Initial user prompt to the AI model.
        params: Parameters to use while initiating the OpenAI chat model. DEFAULT_PARAMS used if not provided.
    """

    def __init__(
        self,
        model: str,
        user_prompt: Optional[str] = None,
        params: Dict[str, float] = DEFAULT_PARAMS,
        semantic_search_query: Optional[str] = None,
    ):
        """Instantiates a new CustomAIChat object.

        Args:
            model: The OpenAI model to use. If not passed, defaults to gpt-3.5-turbo-16k.
            user_prompt: The user prompt to the AI model.
            params: Parameters to use while initiating the OpenAI chat model. DEFAULT_PARAMS used if not provided.
            semantic_search_query: A query string to fetch relevant documents from the database
        """
        self.model = model
        self.messages = [
            {"role": role, "content": content}
            for role, content in [
                ("system", SYSTEM_PROMPT),
                ("user", self._get_doc(semantic_search_query)),
                ("user", user_prompt),
            ]
            if content is not None
        ]
        self.params = params

    @staticmethod
    def _get_doc(semantic_search_query: Optional[str] = None) -> str:
        if semantic_search_query is None:
            return ""
        return _get_relevant_document(semantic_search_query)
    
    @_retry_with_exponential_backoff()
    def __call__(self, user_prompt: str) -> Tuple[str, Dict[str, int]]:
        """Call OpenAI API chat completion endpoint and generate a response.

        Args:
            user_prompt: A string containing user's input prompt.

        Returns:
            A tuple with AI's response message content and the total number of tokens used while generating the response.
        """
        self.messages.append(
            {"role": "user", "content": f"{user_prompt}\n==== YOUR RESPONSE ====\n"}
        )
        prompt_str = "\n\n".join([f"===Role:{m['role']}===\n\nMessage:\n{m['content']}" for m in self.messages])
        logger.info(f"\n\nPrompt to the model: \n\n{prompt_str}")
        
        response = openai.ChatCompletion.create(
            model=self.model,
            messages=self.messages,
            temperature=self.params["temperature"],
        )

        return (
            response["choices"][0]["message"]["content"],
            response["usage"],
        )

# %% ../../nbs/Chat.ipynb 12
class ValidateAndFixResponse:
    """Generates and validates response from OpenAI

    Attributes:
        generate: A callable object for generating responses.
        validate: A callable object for validating responses.
        max_retries: An optional integer specifying the maximum number of attempts to generate and validate a response.
    """

    def __init__(
        self,
        generate: Callable[..., Any],
        validate: Callable[..., Any],
        max_retries: Optional[int] = MAX_RETRIES,
    ):
        self.generate = generate
        self.validate = validate
        self.max_retries = max_retries

    def fix(
        self,
        prompt: str,
        total_usage: List[Dict[str, int]],
        step_name: Optional[str] = None,
        log_dir_path: Optional[str] = None,
        **kwargs: Dict[str, Any],
    ) -> Tuple[str, List[Dict[str, int]]]:
        raise NotImplementedError()

# %% ../../nbs/Chat.ipynb 13
def _save_log_results(
    step_name: str,
    log_dir_path: str,
    messages: List[Dict[str, str]],
    response: str,
    error_str: str,
    retry_cnt: int,
    **kwargs: Dict[str, int],
) -> None:
    if log_dir_path is not None and "attempt" in kwargs:
        step_dir = Path(log_dir_path) / step_name
        step_dir.mkdir(parents=True, exist_ok=True)

        attempt_dir = step_dir / f'attempt_{kwargs["attempt"] + 1}'  # type: ignore
        attempt_dir.mkdir(parents=True, exist_ok=True)

        try_dir = attempt_dir / f"try_{retry_cnt+1}"
        try_dir.mkdir(parents=True, exist_ok=True)

        formatted_msg = "\n".join(
            [f"===={m['role']}====\n\n{m['content']}\n\n" for m in messages]
        )

        with open((try_dir / "input.txt"), "w", encoding="utf-8") as f_input, open(
            (try_dir / "output.txt"), "w", encoding="utf-8"
        ) as f_output, open(
            (try_dir / "errors.txt"), "w", encoding="utf-8"
        ) as f_errors:
            f_input.write(formatted_msg)
            f_output.write(response)
            f_errors.write(error_str)

# %% ../../nbs/Chat.ipynb 15
def _construct_prompt_with_error_msg(
    response: str,
    errors: str,
) -> str:
    """Construct prompt message along with the error message.

    Args:
        prompt: The original prompt string.
        response: The invalid response string from OpenAI.
        errors: The errors which needs to be fixed in the invalid response.

    Returns:
        A string combining the original prompt, invalid response, and the error message.
    """
    prompt_with_errors = (
        f"\n\n==== YOUR RESPONSE (WITH ISSUES) ====\n\n{response}"
        + f"\n\nRead the contents of ==== YOUR RESPONSE (WITH ISSUES) ==== section and fix the below mentioned issues:\n\n{errors}"
    )
    return prompt_with_errors

# %% ../../nbs/Chat.ipynb 17
@patch  # type: ignore
def fix(
    self: ValidateAndFixResponse,
    prompt: str,
    total_usage: List[Dict[str, int]],
    step_name: str,
    output_directory: str,
    **kwargs: Dict[str, Any],
) -> List[Dict[str, int]]:
    """Fix the response from OpenAI until no errors remain or maximum number of attempts is reached.

    Args:
        prompt: The initial prompt string.
        kwargs: Additional keyword arguments to be passed to the validation function.

    Returns:
        str: The generated response that has passed the validation.

    Raises:
        ValueError: If the maximum number of attempts is exceeded and the response has not successfully passed the validation.
    """
    total_tokens_usage: Dict[str, int] = defaultdict(int)
    log_dir_path = Path(output_directory) / LOGS_DIR_NAME
    for i in range(self.max_retries):  # type: ignore
        response, usage = self.generate(prompt)
        total_tokens_usage = add_tokens_usage([total_tokens_usage, usage])
        
        if step_name == STEP_LOG_DIR_NAMES["app"]:
            errors, response = self.validate(response, output_directory, **kwargs)
        else:
            errors = self.validate(response, output_directory, **kwargs)
        error_str = "\n".join(errors)
        _save_log_results(
            step_name,
            str(log_dir_path),
            self.generate.messages,  # type: ignore
            response,
            error_str,
            i,
            **kwargs,
        )
        if len(errors) == 0:
            total_usage.append(total_tokens_usage)
            return total_usage

        self.generate.messages[-1]["content"] = self.generate.messages[-1][ # type: ignore
            "content"
        ].rsplit("==== YOUR RESPONSE ====", 1)[0]
        prompt = _construct_prompt_with_error_msg(response, error_str)
        logger.info(f"Validation failed, trying again...Errors:\n{error_str}")

    total_usage.append(total_tokens_usage)
    
    # we send False to notify the generated code contains bugs
    raise ValueError(total_usage, False)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/App_Description_Validator.ipynb.

# %% auto 0
__all__ = ['logger', 'ERROR_RESPONSE', 'GENERAL_FASTKAFKA_RESPONSE', 'validate_app_description']

# %% ../../nbs/App_Description_Validator.ipynb 1
from typing import *
import time

from yaspin import yaspin

from .._components.logger import get_logger
from .chat import CustomAIChat
from .prompts import APP_VALIDATION_PROMPT
from .constants import INCOMPLETE_DESCRIPTION, DESCRIPTION_EXAMPLE, OpenAIModel

# %% ../../nbs/App_Description_Validator.ipynb 3
logger = get_logger(__name__)

# %% ../../nbs/App_Description_Validator.ipynb 5
ERROR_RESPONSE = "I apologize, but I can only respond to queries related to FastStream code generation. Feel free to ask me about using FastStream, and I'll do my best to help you with that!"
GENERAL_FASTKAFKA_RESPONSE = "Great to see your interest in FastStream! Unfortunately, I can only generate FastStream code and offer assistance in that area. For general information about FastStream, please visit https://faststream.airt.ai/"

# %% ../../nbs/App_Description_Validator.ipynb 6
def validate_app_description(description: str, model: str, total_usage: List[Dict[str, int]]) -> Tuple[str, List[Dict[str, int]]]:
    """Validate the user's application description

    If the description is unrelated to FastStream or contains insensitive/inappropriate language, show an error
    message and exit the program. Otherwise, display the success message in the terminal.

    Args:
        description: User's application description
        
    Raises:
        ValueError: If the application description is invalid
    """
    
    print("✨  Generating a new FastStream application!")
    logger.info("==== App description validation ====")
    with yaspin(
        text="Validating the application description...", color="cyan", spinner="clock"
    ) as sp:
        
        ai = CustomAIChat(user_prompt=APP_VALIDATION_PROMPT, model=model, semantic_search_query="What is FastStream?")
        response, usage = ai(description)
        total_usage.append(usage)
        
        sp.text = ""
        if response == "0":
            raise ValueError(f"✘ Error: Application description validation failed.\n{ERROR_RESPONSE}\n{DESCRIPTION_EXAMPLE}\n\n")
        elif response == "1":
            raise ValueError(f"✘ Error: Application description validation failed.\n\n{GENERAL_FASTKAFKA_RESPONSE}\n{DESCRIPTION_EXAMPLE}\n\n")
        elif response == "2":
            raise ValueError(f"✘ Error: Application description is incomplete.\n\n{INCOMPLETE_DESCRIPTION}\n{DESCRIPTION_EXAMPLE}\n\n")
        else:
            sp.ok(" ✔ Application description validated.")
            return description, total_usage



from pydantic import BaseModel, Field

from faststream import FastStream, Logger
from faststream.kafka import KafkaBroker


class Employee(BaseModel):
    employee_name: str = Field(
        ..., examples=["John Doe"], description="Employee name"
    )
    age: int = Field(
        ..., examples=[30], description="Age"
    )
    location: str = Field(
        ..., examples=["New York"], description="Location"
    )
    experience: int = Field(
        ..., examples=[5], description="Experience in years"
    )


broker = KafkaBroker("localhost:9092")
app = FastStream(broker)


@broker.subscriber("new_joinee")
@broker.publisher("project_team")
@broker.publisher("admin_team")
async def on_new_joinee(msg: Employee, logger: Logger):
    logger.info(msg)
    return msg


# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/Helper.ipynb.

# %% auto 0
__all__ = ['logger', 'examples_delimiter', 'set_cwd', 'set_logger_level', 'retry_on_error', 'ensure_openai_api_key_set',
           'add_tokens_usage', 'get_relevant_prompt_examples', 'strip_white_spaces', 'write_file_contents',
           'read_file_contents', 'mock_openai_create', 'download_and_extract_faststream_archive',
           'validate_python_code']

# %% ../../nbs/Helper.ipynb 1
from typing import *
import os
import re
import functools
import logging
from collections import defaultdict
from tempfile import TemporaryDirectory
from pathlib import Path
from contextlib import contextmanager
import unittest.mock
import zipfile
import importlib.util
import time

import typer
import requests
from langchain.schema.document import Document
from langchain.embeddings import OpenAIEmbeddings
from langchain.vectorstores import FAISS

from .._components.logger import get_logger, set_level
from .._components.logger import suppress_timestamps
from faststream_gen._code_generator.constants import (
    OPENAI_KEY_EMPTY_ERROR,
    OPENAI_KEY_NOT_SET_ERROR,
    TOKEN_TYPES,
    MAX_RESTARTS,
    MAX_RETRIES,
    STEP_LOG_DIR_NAMES,
)
from .._components.package_data import get_root_data_path

# %% ../../nbs/Helper.ipynb 3
logger = get_logger(__name__, level=logging.WARNING)

# %% ../../nbs/Helper.ipynb 5
@contextmanager
def set_cwd(cwd_path: Union[Path, str]) -> Generator:
    """Set the current working directory for the duration of the context manager.

    Args:
        cwd_path: The path to the new working directory.

    !!! note

        The above docstring is autogenerated by docstring-gen library (https://github.com/airtai/docstring-gen)
    """
    cwd_path = Path(cwd_path)
    original_cwd = os.getcwd()
    os.chdir(cwd_path)

    try:
        yield
    finally:
        os.chdir(original_cwd)

# %% ../../nbs/Helper.ipynb 7
def set_logger_level(func: Callable[..., Any]) -> Callable[..., Any]:
    """Decorator to set the logger level based on verbosity.

    Args:
        func: The function to be decorated.

    Returns:
        The decorated function.
    """

    @functools.wraps(func)
    def wrapper_decorator(*args, **kwargs): # type: ignore
        if ("verbose" in kwargs) and kwargs["verbose"]:
            set_level(logging.INFO)
        else:
            set_level(logging.WARNING)
        return func(*args, **kwargs)

    return wrapper_decorator

# %% ../../nbs/Helper.ipynb 10
def retry_on_error(max_retries: int = MAX_RESTARTS, delay: int = 1):  # type: ignore
    def decorator(func):  # type: ignore
        def wrapper(*args, **kwargs):  # type: ignore
            for i in range(max_retries):
                try:
                    kwargs["attempt"] = i
                    return func(*args, **kwargs)
                except ValueError as e:
                    # Log the error here
                    logger.info(f"Attempt {i} failed. Restarting step.")
                    time.sleep(delay)
                    # Capture exception details here
                    last_exception = e
            return last_exception.args[0], last_exception.args[1]
        return wrapper

    return decorator

# %% ../../nbs/Helper.ipynb 13
def ensure_openai_api_key_set() -> None:
    """Ensure the 'OPENAI_API_KEY' environment variable is set and is not empty.

    Raises:
        KeyError: If the 'OPENAI_API_KEY' environment variable is not found.
        ValueError: If the 'OPENAI_API_KEY' environment variable is found but its value is empty.
    """
    try:
        openai_api_key = os.environ["OPENAI_API_KEY"]
        if openai_api_key == "":
            raise ValueError(OPENAI_KEY_EMPTY_ERROR)
    except KeyError:
        raise KeyError(OPENAI_KEY_NOT_SET_ERROR)

# %% ../../nbs/Helper.ipynb 17
def add_tokens_usage(usage_list: List[Dict[str, int]]) -> Dict[str, int]:
    """Add list of OpenAI "usage" dictionaries by categories defined in TOKEN_TYPES (prompt_tokens, completion_tokens and total_tokens).

    Args:
        usage_list: List of OpenAI "usage" dictionaries


    Returns:
        Dict[str, int]: Dictionary where the keys are TOKEN_TYPES and their values are the sum of OpenAI "usage" dictionaries
    """
    added_tokens: Dict[str, int] = defaultdict(int)
    for usage in usage_list:
        for token_type in TOKEN_TYPES:
            added_tokens[token_type] += usage[token_type]
            
    return added_tokens

# %% ../../nbs/Helper.ipynb 20
examples_delimiter = {
    "description": {
        "start": "==== description.txt starts ====",
        "end": "==== description.txt ends ====",
    },
    "skeleton": {
        "start": "==== app_skeleton.py starts ====",
        "end": "==== app_skeleton.py ends ====",
    },
    "app": {
        "start": "==== app.py starts ====",
        "end": "==== app.py ends ====",
    },
    "test_app": {
        "start": "==== test_app.py starts ====",
        "end": "==== test_app.py ends ====",
    },
}


def _split_text(text: str, delimiter: Dict[str, str]) -> str:
    return text.split(delimiter["start"])[-1].split(delimiter["end"])[0]


def _format_examples(parent_docs_str: List[str]) -> Dict[str, str]:
    """Format and extract examples from parent document.

    Args:
        parent_docs_str (List[str]): A list of parent document strings containing example sections.

    Returns:
        Dict[str, List[str]]: A dictionary with sections as keys and lists of formatted examples as values.
    """
    ret_val = {"description_to_skeleton": "", "skeleton_to_app_and_test": ""}
    for d in parent_docs_str:
        description = _split_text(d, examples_delimiter["description"])
        skeleton = _split_text(d, examples_delimiter["skeleton"])
        app = _split_text(d, examples_delimiter["app"])
        test_app = _split_text(d, examples_delimiter["test_app"])

        ret_val[
            "description_to_skeleton"
        ] += f"\n==== EXAMPLE APP DESCRIPTION ====\n{description}\n\n==== YOUR RESPONSE ====\n\n{skeleton}"
        ret_val[
            "skeleton_to_app_and_test"
        ] += f"\n==== EXAMPLE APP DESCRIPTION ====\n{description}\n\n==== EXAMPLE APP SKELETON ====\n{skeleton}\n==== YOUR RESPONSE ====\n\n### application.py ###\n{app}\n### test.py ###\n{test_app}"

    return ret_val

# %% ../../nbs/Helper.ipynb 22
def get_relevant_prompt_examples(query: str) -> Dict[str, str]:
    """Load the vector database and retrieve the most relevant examples based on the given query for each step.

    Args:
        query: The query for relevance-based document retrieval.

    Returns:
        The dictionary of the most relevant examples for each step.
    """
    db_path = get_root_data_path() / "examples"
    db = FAISS.load_local(db_path, OpenAIEmbeddings()) # type: ignore
    results = db.similarity_search(query, k=3, fetch_k=5)
    results_page_content = [r.page_content for r in results]
    prompt_examples = _format_examples(results_page_content)
    return prompt_examples

# %% ../../nbs/Helper.ipynb 24
def strip_white_spaces(description: str) -> str:
    """Remove and strip excess whitespaces from a given description

    Args:
        description: The description string to be processed.

    Returns:
        The cleaned description string.
    """
    pattern = re.compile(r"\s+")
    return pattern.sub(" ", description).strip()

# %% ../../nbs/Helper.ipynb 26
def write_file_contents(output_file: str, contents: str) -> None:
    """Write the given contents to the specified output file.

    Args:
        output_file: The path to the output file where the contents will be written.
        contents: The contents to be written to the output file.

    Raises:
        OSError: If there is an issue while attempting to save the file.
    """
    try:
        Path(output_file).parent.mkdir(parents=True, exist_ok=True)

        with open(output_file, "w", encoding="utf-8") as f:
            f.write(contents)

    except OSError as e:
        raise OSError(
            f"Error: Failed to save file at '{output_file}' due to: '{e}'. Please ensure that the specified 'output_path' is valid and that you have the necessary permissions to write files to it."
        )

# %% ../../nbs/Helper.ipynb 28
def read_file_contents(output_file: str) -> str:
    """Read and return the contents from the specified file.

    Args:
        output_file: The path to the file to be read.

    Returns:
        The contents of the file as string.

    Raises:
        FileNotFoundError: If the specified file does not exist.
    """
    try:
        with open(output_file, "r", encoding="utf-8") as f:
            contents = f.read()
        return contents
    except FileNotFoundError:
        raise FileNotFoundError(
            f"Error: The file '{output_file}' does not exist. Please ensure that the specified 'output_path' is valid and that you have the necessary permissions to access it."
        )

# %% ../../nbs/Helper.ipynb 31
@contextmanager
def mock_openai_create(test_response: str) -> Generator[None, None, None]:
    mock_choices = {
        "choices": [{"message": {"content": test_response}}],
        "usage": { 
            "prompt_tokens": 129,
            "completion_tokens": 1,
            "total_tokens": 130
        },
    }

    with unittest.mock.patch("openai.ChatCompletion") as mock:
        mock.create.return_value = mock_choices
        yield

# %% ../../nbs/Helper.ipynb 33
def _fetch_content(url: str) -> requests.models.Response: # type: ignore
    """Fetch content from a URL using an HTTP GET request.

    Args:
        url (str): The URL to fetch content from.

    Returns:
        Response: The response object containing the content and HTTP status.

    Raises:
        requests.exceptions.Timeout: If the request times out.
        requests.exceptions.RequestException: If an error occurs during the request.
    """
    attempt = 0
    while attempt < 4:
        try:
            response = requests.get(url, timeout=50)
            response.raise_for_status()  # Raises an exception for HTTP errors
            return response
        except requests.exceptions.Timeout:
            if attempt == 3:  # If this was the fourth attempt, raise the Timeout exception
                raise requests.exceptions.Timeout(
                    "Request timed out. Please check your internet connection or try again later."
                )
            time.sleep(1)  # Sleep for one second before retrying
            attempt += 1
        except requests.exceptions.RequestException as e:
            raise requests.exceptions.RequestException(f"An error occurred: {e}")

# %% ../../nbs/Helper.ipynb 35
@contextmanager
def download_and_extract_faststream_archive(url: str) -> Generator[Path, None, None]:
    with TemporaryDirectory() as d:
        try:
            input_path = Path(f"{d}/archive.zip")
            extrated_path = Path(f"{d}/extrated_path")
            extrated_path.mkdir(parents=True, exist_ok=True)

            response = _fetch_content(url)

            with open(input_path, "wb") as f:
                f.write(response.content)

            with zipfile.ZipFile(input_path, "r") as zip_ref:
                for member in zip_ref.namelist():
                    zip_ref.extract(member, extrated_path)

            yield extrated_path

        except Exception as e:
            fg = typer.colors.RED
            typer.secho(f"Unexpected internal error: {e}", err=True, fg=fg)
            raise typer.Exit(code=1)

# %% ../../nbs/Helper.ipynb 37
def validate_python_code(file_name: str, **kwargs: Dict[str, Any]) -> List[str]:
    """Validate and report errors in the provided Python code.

    Args:
        file_name: Python file to validate

    Returns:
        A list of error messages encountered during validation. If no errors occur, an empty list is returned.
    """
    try:
        # Import the module using importlib
        spec = importlib.util.spec_from_file_location("tmp_module", file_name)
        module = importlib.util.module_from_spec(spec)  # type: ignore
        spec.loader.exec_module(module)  # type: ignore

    except Exception as e:
        return [f"{type(e).__name__}: {e}"]

    return []



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/Constants.ipynb.

# %% auto 0
__all__ = ['APPLICATION_FILE_PATH', 'TEST_FILE_PATH', 'TOML_FILE_NAME', 'LOGS_DIR_NAME', 'STEP_LOG_DIR_NAMES', 'DEFAULT_PARAMS',
           'MAX_RETRIES', 'MAX_RESTARTS', 'MAX_ASYNC_SPEC_RETRIES', 'TOKEN_TYPES', 'MODEL_PRICING',
           'OPENAI_KEY_EMPTY_ERROR', 'OPENAI_KEY_NOT_SET_ERROR', 'EMPTY_DESCRIPTION_ERROR', 'INCOMPLETE_DESCRIPTION',
           'DESCRIPTION_EXAMPLE', 'MAX_NUM_FIXES_MSG', 'INCOMPLETE_APP_ERROR_MSG', 'FASTSTREAM_REPO_ZIP_URL',
           'FASTSTREAM_DOCS_DIR_SUFFIX', 'FASTSTREAM_EXAMPLES_DIR_SUFFIX', 'FASTSTREAM_EXAMPLE_FILES',
           'FASTSTREAM_TMP_DIR_PREFIX', 'FASTSTREAM_DIR_TO_EXCLUDE', 'FASTSTREAM_TEMPLATE_ZIP_URL',
           'FASTSTREAM_TEMPLATE_DIR_SUFFIX', 'OpenAIModel']

# %% ../../nbs/Constants.ipynb 2
APPLICATION_FILE_PATH = "app/application.py"
TEST_FILE_PATH = "tests/test_application.py"
TOML_FILE_NAME = "pyproject.toml"
LOGS_DIR_NAME = "_faststream_gen_logs"

STEP_LOG_DIR_NAMES = {
    "skeleton": "app-skeleton-generation-logs",
    "app": "app-and-test-generation-logs",
    "requirements": "requirements-generation-logs",
}

# %% ../../nbs/Constants.ipynb 4
DEFAULT_PARAMS = {
    "temperature": 0.7,
}

MAX_RETRIES = 3
MAX_RESTARTS = 3
MAX_ASYNC_SPEC_RETRIES = 3


from enum import Enum
class OpenAIModel(str, Enum):
    gpt3 = "gpt-3.5-turbo-16k"
    gpt4 = "gpt-4"


# %% ../../nbs/Constants.ipynb 7
TOKEN_TYPES = ["prompt_tokens", "completion_tokens", "total_tokens"]

MODEL_PRICING = {
    OpenAIModel.gpt4.value: {
        "input": 0.03,
        "output": 0.06
    },
    OpenAIModel.gpt3.value: {
        "input": 0.003,
        "output": 0.004
    },
}

# %% ../../nbs/Constants.ipynb 9
OPENAI_KEY_EMPTY_ERROR = "Error: OPENAI_API_KEY cannot be empty. Please set a valid OpenAI API key in OPENAI_API_KEY environment variable and try again.\nYou can generate API keys in the OpenAI web interface. See https://platform.openai.com/account/api-keys for details."
OPENAI_KEY_NOT_SET_ERROR = "Error: OPENAI_API_KEY not found in environment variables. Set a valid OpenAI API key in OPENAI_API_KEY environment variable and try again. You can generate API keys in the OpenAI web interface. See https://platform.openai.com/account/api-keys for details."

EMPTY_DESCRIPTION_ERROR = "Error: you need to provide the application description by providing it with the command line argument or by providing it within a textual file wit the --input_file argument."



INCOMPLETE_DESCRIPTION = """Please check if your application description is missing some crucial information:
- Description of the messages that will be produced or consumed
- At least one topic
- The business logic to implement while consuming or producing the messages
"""
DESCRIPTION_EXAMPLE = """
If you're unsure about how to construct the app description, consider the following example for guidance

APPLICATION DESCRIPTION EXAMPLE:
Create a FastStream application using localhost broker for testing and use the default port number. 
It should consume messages from the 'input_data' topic, where each message is a JSON encoded object containing a single attribute: 'data'. 
For each consumed message, create a new message object and increment the value of the data attribute by 1. Finally, send the modified message to the 'output_data' topic.
"""

MAX_NUM_FIXES_MSG = "Maximum number of retries"

INCOMPLETE_APP_ERROR_MSG = """Apologies, we couldn't generate a working application and test code from your application description.

Please run the following command to start manual debugging:"""

# %% ../../nbs/Constants.ipynb 11
FASTSTREAM_REPO_ZIP_URL = "http://github.com/airtai/faststream/archive/main.zip"
FASTSTREAM_DOCS_DIR_SUFFIX = "faststream-main/.faststream_gen"
FASTSTREAM_EXAMPLES_DIR_SUFFIX = "faststream-main/faststream_gen_examples"
FASTSTREAM_EXAMPLE_FILES = ['description.txt', 'app_skeleton.py', 'app.py', 'test_app.py']
FASTSTREAM_TMP_DIR_PREFIX = "appended_examples"
FASTSTREAM_DIR_TO_EXCLUDE = "api"

# %% ../../nbs/Constants.ipynb 13
FASTSTREAM_TEMPLATE_ZIP_URL = "http://github.com/airtai/faststream-template/archive/main.zip"
FASTSTREAM_TEMPLATE_DIR_SUFFIX = "faststream-template-main"
